"""A translation of a Scheme interpreter in Python, modeling the
metacircular evaluator in Ch. 4 of The Structure and Interpretation of
Computer Programs.  (http://www-mitpress.mit.edu/sicp/)

AUTHOR:

    Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

SYNOPSIS:

    It's Scheme.  *grin*

    Scheme is a variant of Lisp, and it's used in many introductory CS
    courses.  It is pretty powerful despite its deceptive size.  For
    more information on Scheme, visit the Scheme home page:

        http://www.swiss.ai.mit.edu/projects/scheme

BUGS:

    One major deficiency of this implementation is that it doesn't
    support tail recursion.  I don't believe I can implement this
    properly without fiddling around with the 'compiler' package and
    primitive bytecodes.  I'll read up on SICP Chapter 5.5 and see if
    this is feasible.

    August 26, 2004: this is not true; I CAN get real tail call
    elimination working, but it'll take CPS and trampolining to get it
    working right.  I'll first modify the interpreter into CPS form,
    and then apply trampolining from there.... ok, done!
    
    Another is the broken implementation of pairs --- I try to use
    Python's list structure to simulate pair structure, but this
    implementation has flaws.  For example, it's a known bug that:

       (define x 3)
       (define y '(4))
       (eq? (cdr (cons x y) y))

    doesn't work the way it should.

    In a previous version of PyScheme, CONSing two atoms together was
    prohibited entirely.  Howver, Seth David Schoen suggested a
    workaround, which I've implemented.  We can now do

        (cons 1 2)

    and get something that looks right.  [note: august 26: this is
    something I need to look at again: my pair implementation is
    breaking badly.]


MISCELLANEOUS:

    This source code is for the public domain; I'm writing it just for
    the fun of it.  This is certainly not meant to be efficient or
    even useful.  *grin*

    My future goals is to make this usable enough to run something
    like the explicit-register simulator and compiler in SICP Ch. 5.

"""


from __future__ import nested_scopes

import types, math, sys, pprint, exceptions

"""I need to use the operator module occasionally, but I've also
defined a 'operator' selector function in this program.  To avoid
confusion, I'll rename the 'operator' module as 'python_operator'."""
import operator as python_operator

import string
from symbol import Symbol, true, false, isSymbol
from env import *
from pair import *
from parser import parse
import pogo

from error import SchemeError


"""We need to save Python's apply() function, since we use it later on
to evaluate primitive procedure calls.  For symmetry, I'm also saving
Python's eval() function within evalInUnderlyingPython()."""
evalInUnderlyingPython = eval
applyInUnderlyingPython = apply


def identity(val):
    """The identity function."""
    return val



def apply_cont(cont, val):
    return pogo.bounce(cont, val)



######################################################################
## The heart of the interpreter is eval/apply.
##
## Some notes here: I'm using trampolined style to get around Python's
## lack of tail recursion.

def eval(exp, env, cont=identity):
    landing_cont = lambda v: pogo.land(v)
    return cont(pogo.pogo_start(teval, exp, env, landing_cont))


def teval(exp, env, cont):
    """Evaluates an expression 'exp' in an environment 'env'.

    Exercise 4.3 asks us to rewrite this in a more natural
    data-directed manner.
    """
    if isSelfEvaluating(exp):
        return apply_cont(cont, exp)
    if isVariable(exp):
        return apply_cont(cont, lookupVariableValue(exp, env))
    if isQuoted(exp):
        return evalQuoted(exp, env, cont)
    if isAssignment(exp):
        return evalAssignment(exp, env, cont)
    if isDefinition(exp):
        return evalDefinition(exp, env, cont)
    if isLet(exp):
        return teval(letToApplication(exp), env, cont)
    if isIf(exp):
        return evalIf(exp, env, cont)
    if isLambda(exp):
        return apply_cont(cont, makeProcedure(lambdaParameters(exp),
                                              lambdaBody(exp),
                                              env))
    if isBegin(exp):
        return evalSequence(beginActions(exp), env, cont)
    if isCond(exp):
        return teval(condToIf(exp), env, cont)
    if isLoad(exp):
        return load(exp, env, cont)
    if isCallcc(exp):      ## Hmmm.. this doesn't need to be a special form, does it?
        return evalCallcc(exp, env, cont)
    if isApplication(exp):
        return evalApplication(exp, env, cont)
    raise SchemeError, "Unknown expression type -- eval " + exp



def apply(procedure, arguments, cont=identity):
    """Applies a procedure on a list of arguments."""
    if isPrimitiveProcedure(procedure):
        return applyPrimitiveProcedure(procedure, arguments, cont)
    elif isContinuationProcedure(procedure):
        return applyContinuationProcedure(procedure, arguments)
    if isCompoundProcedure(procedure):
        newEnv = extendEnvironment(pairsToPythonList(procedureParameters(procedure)),
                                   pairsToPythonList(arguments),
                                   procedureEnvironment(procedure))
        return evalSequence(procedureBody(procedure), newEnv, cont)
    raise SchemeError, "Unknown procedure type -- apply " + str(procedure)



def evalRands(exps, env, cont=identity):
    """Given a list of expressions, returns a new list containing the
    values of evaluating each on of them.  If the continuation is
    given, then calls cont() on the evaluated operands instead."""
    def c1(head_val):
        def c2(tail_val):
            return apply_cont(cont, schemeCons(head_val, tail_val))
        return evalRands(restOperands(exps), env, c2)
    if isNoOperands(exps):
        return apply_cont(cont, schemeList())
    return teval(firstOperand(exps), env, c1)


def evalIf(exp, env, cont=identity):
    def c(predicate_val):
        if isTrue(predicate_val):
            return teval(ifConsequent(exp), env, cont)
        else:
            return teval(ifAlternative(exp), env, cont)
    return teval(ifPredicate(exp), env, c)


def evalSequence(exps, env, cont=identity):
    def c(val):
        return evalSequence(restExps(exps), env, cont)
    if isLastExp(exps):
        return teval(firstExp(exps), env, cont)
    else:
        return teval(firstExp(exps), env, c)


def evalAssignment(exp, env, cont=identity):
    def c(val):
        setVariableValue(assignmentVariable(exp), val, env)
        return apply_cont(cont, Symbol("ok"))
    return teval(assignmentValue(exp), env, c)


def evalDefinition(exp, env, cont=identity):
    def c(val):
        defineVariable(definitionVariable(exp), val, env)
        return apply_cont(cont, Symbol("ok"))
    return teval(definitionValue(exp), env, c)


def evalApplication(exp, env, cont=identity):
    def c1(operator_val):
        def c2(operands_val):
            return apply(operator_val, operands_val, cont)
        return evalRands(operands(exp), env, c2)
    return teval(operator(exp), env, c1)


def evalCallcc(exp, env, cont=identity):
    def c1(lambda_val):
        def c2(val):
            return cont(val)
        cont_procedure = makeContinuationProcedure(c2)
        return apply(lambda_val, schemeList(cont_procedure), cont)
    return teval(callccBody(exp), env, c1)


######################################################################
                     

def isSelfEvaluating(exp):
    if isNumber(exp): return 1
    if isString(exp): return 1
    return 0


def isNumber(x):
    return type(x) in (types.IntType, types.FloatType, types.LongType)


def isString(x):
    return type(x) == types.StringType


def isVariable(exp):
    return isSymbol(exp)


def isQuoted(exp):
    return isTaggedList(exp, Symbol("quote"))

def isUnquoted(exp):
    return isTaggedList(exp, Symbol("unquote"))


def textOfQuotation(exp):
    return schemeCadr(exp)


def textOfUnquotation(exp):
    return schemeCadr(exp)


def isTaggedList(exp, tag):
    if isList(exp) and schemeLength(exp) > 0:
        return schemeCar(exp) == tag
    return 0



######################################################################

def evalQuoted(exp, env, cont=identity):
    """Returns a quoted expression, using deepQuotedEval to look for
    UNQUOTE.

    Consequently, quoted elements are properly turned into cons pairs.
    """
    return deepQuotedEval(textOfQuotation(exp), env, cont)


## FIXME: not quite right: I should be doing unquotation only on
## quasiquotes, not on just any quote... Oops.
def deepQuotedEval(exp, env, cont=identity):
    if not isList(exp):
        return apply_cont(cont, exp)
    if isNull(exp):
        return apply_cont(cont, exp)
    if isUnquoted(exp):
        return teval(textOfUnquotation(exp), env, cont)
    ## FIXME: this last clause is not quite right as far as continuations are
    ## concerned.
    return apply_cont(cont, schemeCons(deepQuotedEval(schemeCar(exp), env),
                                       deepQuotedEval(schemeCdr(exp), env)))

######################################################################


def isCallcc(exp):
    return (isTaggedList(exp, Symbol("call/cc")) or
            isTaggedList(exp, Symbol("call-with-current-continuation")))

def callccBody(exp):
    return schemeCadr(exp)


######################################################################

def isLet(exp):
    return isTaggedList(exp, Symbol("let"))

def letBindings(exp):
    return schemeCadr(exp)

def letBindingVariables(bindings):
    if isNull(bindings): return schemeList()
    return schemeCons(schemeCar(schemeCar(bindings)),
                      letBindingVariables(schemeCdr(bindings)))

def letBindingValues(bindings):
    if isNull(bindings): return schemeList()
    return schemeCons(schemeCadr(schemeCar(bindings)),
                      letBindingValues(schemeCdr(bindings)))

def letBody(exp):
    return schemeCdr(schemeCdr(exp))

def letToApplication(exp):
    bindings = letBindings(exp)
    return schemeCons(makeLambda(letBindingVariables(bindings),
                                 letBody(exp)),
                      letBindingValues(bindings))

######################################################################


def isAssignment(exp):
    """Assignments have the form (set! <var> <value>)."""
    return isTaggedList(exp, Symbol("set!"))


def assignmentVariable(exp):
    return schemeCadr(exp)


def assignmentValue(exp):
    return schemeCaddr(exp)

######################################################################

def isDefinition(exp):
    return isTaggedList(exp, Symbol("define"))


def definitionVariable(exp):
    if isSymbol(schemeCadr(exp)):
        return schemeCadr(exp)
    return schemeCar(schemeCadr(exp))


def definitionValue(exp):
    if isSymbol(schemeCadr(exp)):
        return schemeCaddr(exp)
    return makeLambda(schemeCdr(schemeCadr(exp)),
                      schemeCddr(exp))


######################################################################

def isLambda(exp):
    return isTaggedList(exp, Symbol("lambda"))


def lambdaParameters(exp):
    return schemeCadr(exp)


def lambdaBody(exp):
    return schemeCddr(exp)


def makeLambda(parameters, body):
    if schemeCount(Symbol("."), parameters) > 1:
        raise SchemeError, \
              "Parameter list can have at most one vararg parameter -- LAMBDA"
    return schemeAppend(schemeList(Symbol("lambda"), parameters),
                        body)


######################################################################


def isIf(exp):
    return isTaggedList(exp, Symbol("if"))


def ifPredicate(exp):
    return schemeCadr(exp)


def ifConsequent(exp):
    return schemeCaddr(exp)


def ifAlternative(exp):
    if not isNull(schemeCdddr(exp)):
        return schemeCadddr(exp)
    return false


def makeIf(predicate, consequent, alternative):
    return schemeList(Symbol("if"), predicate, consequent, alternative)


######################################################################

def isBegin(exp):
    return isTaggedList(exp, Symbol("begin"))


def beginActions(exp):
    return schemeCdr(exp)


def isLastExp(seq):
    return isNull(schemeCdr(seq))


def firstExp(seq):
    return schemeCar(seq)


def restExps(seq):
    return schemeCdr(seq)


def sequenceToExp(seq):
    if isNull(seq): return seq
    if isLastExp(seq): return firstExp(seq)
    return makeBegin(seq)


def makeBegin(seq):
    return schemeCons(Symbol("begin"), seq)


######################################################################


def isApplication(exp):
    return isList(exp)


def operator(exp):
    if schemeLength(exp) == 0:
        raise SchemeError, \
              "No operator given for procedure application -- OPERATOR"
    return schemeCar(exp)


def operands(exp):
    return schemeCdr(exp)


def isNoOperands(ops):
    return isNull(ops)


def firstOperand(ops):
    return schemeCar(ops)


def restOperands(ops):
    return schemeCdr(ops)


######################################################################


def isCond(exp):
    return isTaggedList(exp, Symbol("cond"))


def condClauses(exp):
    return schemeCdr(exp)


def isCondElseClause(clause):
    return condPredicate(clause) == Symbol("else")


def condPredicate(clause):
    return schemeCar(clause)


def condActions(clause):
    return schemeCdr(clause)


def condToIf(exp):
    return expandClauses(condClauses(exp))


def expandClauses(clauses):
    if isNull(clauses):
        return false
    first = schemeCar(clauses)
    rest = schemeCdr(clauses)
    if isCondElseClause(first):
        if isNull(rest):
            return sequenceToExp(condActions(first))
        raise SchemeError, "else clause isn't last -- condToIf" + clauses
    return makeIf(condPredicate(first),
                  sequenceToExp(condActions(first)),
                  expandClauses(rest))
                                

def isTrue(x):
    return x != false


def isFalse(x):
    return x == false

######################################################################

def isLoad(x):
    return isTaggedList(x, Symbol("load"))

def loadFilename(exp):
    return schemeCadr(exp)

def load(exp, env, cont=identity):
    try:
        f = open(loadFilename(exp))
        text = "(begin \n%s\n 'ok)" % f.read()
        return teval(parse(text), env, cont)
    except IOError, e:
        raise SchemeError, "LOAD error -- %s" % str(e)
    except TypeError, e:
        raise SchemeError, "LOAD error -- argument must be a string"
        


######################################################################


def makeProcedure(parameters, body, env):
    return schemeList(Symbol("procedure"), parameters, body, env)

def isCompoundProcedure(p):
    return isTaggedList(p, Symbol("procedure"))

def procedureParameters(p):
    return schemeCadr(p)

def procedureBody(p):
    return schemeCaddr(p)

def procedureEnvironment(p):
    return schemeCadddr(p)



######################################################################
"""Here's where we define the primitives of Scheme."""
def isPrimitiveProcedure(proc):
    return isTaggedList(proc, Symbol("primitive"))


def primitiveImplementation(proc):
    return schemeCadr(proc)


def makePrimitiveProcedure(proc):
    return schemeList(Symbol("primitive"), proc)


## We have to tag a continuation slightly different: applying a
## continuation abandons the current computation.
def isContinuationProcedure(proc):
    return isTaggedList(proc, Symbol("continuation"))


def continuationImplementation(proc):
    return schemeCadr(proc)


def makeContinuationProcedure(proc):
    return schemeList(Symbol("continuation"), proc)


######################################################################



def schemeApply(procedure, arguments):
    return apply(procedure, pairsToPythonList(arguments))

def schemeAdd(*numbers):
    if len(numbers) == 0: return 0
    return reduce(python_operator.add, numbers)

def schemeSubtract(*numbers):
    if len(numbers) == 0:
        raise SchemeError, "primitive '-' --- expects 1 or more arguments"
    return reduce(python_operator.sub, numbers)

def schemeMultiply(*numbers):
    if len(numbers) == 0: return 1
    return reduce(python_operator.mul, numbers)

def schemeDivide(*numbers):
    if len(numbers) == 0:
        raise SchemeError, "primitive '/' --- expects 1 or more arguments"
    return reduce(python_operator.div, numbers)

def schemeRemainder(a, b): return a % b
def schemeQuotient(a, b): return int(a / b)
def schemeSqrt(a): return math.sqrt(a)
def schemeFloor(a): return int(math.floor(a))
def schemeCeiling(a): return int(math.ceil(a))

def schemeNumericalEq(a, b):
    """Here, we need to make sure that we're returning the metacircular
    values of true or false."""
    return schemeBooleanize(a == b)

def schemeLessThan(a, b): return schemeBooleanize(a < b)
def schemeLessThanOrEquals(a, b): return schemeBooleanize(a <= b)
def schemeGreaterThan(a, b): return schemeBooleanize(a > b)
def schemeGreaterThanOrEquals(a, b): return schemeBooleanize(a >= b)

def schemeEqQuestion(a, b):
    if type(a) in (types.IntType, types.FloatType, types.LongType):
        return schemeBooleanize(a == b)
    return schemeBooleanize(a is b)

def schemeEqualQuestion(a, b): return schemeBooleanize(a == b)
def schemeListQuestion(a): return schemeBooleanize(isList(a))
def schemeNullQuestion(a): return schemeBooleanize(isNull(a))

def schemeDisplay(a):
    sys.stdout.write(expressionToString(a))
    return Symbol("ok")
def schemeNewline():
    sys.stdout.write("\n")
    return Symbol("ok")

def schemeStringToSymbol(mystr): return Symbol(mystr)
def schemeSymbolToString(mysymbol): return str(mysymbol)
def schemeNumberToString(mynumber): return str(mynumber)

def schemeStringAppend(*strings):
    if len(strings) == 0: return ""
    return reduce(python_operator.add, strings)

def schemeNot(x):
    if isTrue(x): return false
    return true

def schemeQuit():
    sys.stdout.write("BYE\n")
    sys.exit(0)

def schemeBooleanize(x):
    """Translates what Python considers true and false to the Scheme
    symbols "true" or "false"."""
    if x: return true
    return false



def setupEnvironment():
    PRIMITIVE_PROCEDURES = [
        [Symbol("car"), schemeCar],
        [Symbol("cdr"), schemeCdr],
        [Symbol("cons"), schemeCons],
        [Symbol("append"), schemeAppend],
        [Symbol("list"), schemeList],
        [Symbol("set-car!"), schemeSetCarBang],
        [Symbol("set-cdr!"), schemeSetCdrBang],
        [Symbol("apply"), schemeApply],
        [Symbol("+"), schemeAdd],
        [Symbol("-"), schemeSubtract],
        [Symbol("*"), schemeMultiply],
        [Symbol("/"), schemeDivide],
        [Symbol("remainder"), schemeRemainder],
        [Symbol("quotient"), schemeQuotient],
        [Symbol("sqrt"), schemeSqrt],
        [Symbol("floor"), schemeFloor],
        [Symbol("ceiling"), schemeCeiling],
        [Symbol("="), schemeNumericalEq],
        [Symbol("<"), schemeLessThan],
        [Symbol("<="), schemeLessThanOrEquals],
        [Symbol(">"), schemeGreaterThan],
        [Symbol(">="), schemeGreaterThanOrEquals],
        [Symbol("eq?"), schemeEqQuestion],
        [Symbol("equal?"), schemeEqualQuestion],
        [Symbol("list?"), schemeListQuestion],
        [Symbol("pair?"), schemePairQuestion],
        [Symbol("null?"), schemeNullQuestion],
        [Symbol("display"), schemeDisplay],
        [Symbol("newline"), schemeNewline],
        [Symbol("not"), schemeNot],
        [Symbol("string->symbol"), schemeStringToSymbol],
        [Symbol("symbol->string"), schemeSymbolToString],
        [Symbol("number->string"), schemeNumberToString],
        [Symbol("string-append"), schemeStringAppend],
        [Symbol("quit"), schemeQuit],
        [Symbol("exit"), schemeQuit],
        ]

    PRIMITIVE_PROCEDURE_NAMES = map(lambda x: x[0], PRIMITIVE_PROCEDURES)

    PRIMITIVE_PROCEDURE_OBJECTS = map(lambda proc:
                                      [Symbol("primitive"), proc[1]],
                                      PRIMITIVE_PROCEDURES)


    initial_environment = extendEnvironment(PRIMITIVE_PROCEDURE_NAMES,
                                            PRIMITIVE_PROCEDURE_OBJECTS,
                                            THE_EMPTY_ENVIRONMENT)
    defineVariable(Symbol("#t"), Symbol("#t"), initial_environment)
    defineVariable(Symbol("#f"), Symbol("#f"), initial_environment)
    return initial_environment


def applyPrimitiveProcedure(proc, args, cont):
    try:
        return apply_cont(cont,
                          applyInUnderlyingPython(primitiveImplementation(proc),
                                                  pairsToPythonList(args)))
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
        raise SchemeError, e



def applyContinuationProcedure(proc, args):
    try:
        return applyInUnderlyingPython(continuationImplementation(proc),
                                       pairsToPythonList(args))
    except Exception, e:
        if isinstance(e, exceptions.SystemExit): raise e
        raise SchemeError, e

######################################################################

def expressionToString(expr):
    """Given a Scheme expression, returns a string that tries to nicely
    print it out."""
    if not isPair(expr):
        return str(expr)
    if isCompoundProcedure(expr):
        return '(%s %s %s <procedure-env>)' % \
               (Symbol('compound-procedure'),
                expressionToString(procedureParameters(expr)),
                expressionToString(procedureBody(expr)))
    pieces = []
    while expr:
        pieces.append(expressionToString(schemeCar(expr)))
        if isDottedPair(expr):
            pieces.append(".")
            pieces.append(expressionToString(schemeCdr(expr)))
            break
        expr = schemeCdr(expr)
    return '(' + ' '.join(pieces) + ')'


######################################################################

## def test1():
##     """This is a small test harness that tries to at least touch the
##     functionality that the interpreter provides.  I should formally
##     use unit testing for this when I get the chance."""
##     from parser import parse
##     global_environment = setupEnvironment()
##     commands = parse("""(
##     (+ 3 4)
##     (define pi 3.1415926)
##     pi
##     (* 4 2)
##     (car (cdr (cons 3 (cons 4 '(5)))))
##     ((lambda (x) (* x x)) 42)
##     ((lambda (x y z) (+ (* x y) z)) 3 4 5)
##     (define square (lambda (x) (* x x)))
##     (square 42)
##     (define (hypotenuse a b) (sqrt (+ (square a) (square b))))
##     (hypotenuse 3 4)
##     (define (factorial x)
##     (if (= x 0) 1
##     (* x (factorial (- x 1)))))
##     (factorial 0)
##     (factorial 1)
##     (factorial 2)
##     (factorial 3)
##     (begin
##     (display 1)
##     (display 2)
##     (display 3)
##     (newline)
##     'is-this-working?)
##     (define (even? x) (= (remainder x 2) 0))
##     (define (odd? x) (not (even? x)))
##     (cond ((even? 3) "not-ok")
##           ((odd? 2) "still-not-ok")
##           (#t     "ok"))
##     (even? (square 42))
##     (odd? (square 42))

##     (append '(1 2 3) '(4 5 6))
##     (list '(1 2 3) 4)
##     '(hello do you see how ,pi is unquoted?)
##     '(this (pi (,pi) is deep))
##     (define (mylist . elements) elements)
##     (append (mylist 1 2 3 4 5)
##             (apply mylist '(6 7 8 9 10)))
##     )""")
##     for c in commands:
##         pprint.pprint(c)
##         print expressionToString(eval(c, global_environment))
##         print
##     return global_environment


def repl():
    from parser import parse, ParserError
    from prompt import Prompt
    global_environment = setupEnvironment()
    def promptCallback(s):
        try:
            print expressionToString(eval(parse(s),
                                          global_environment,
                                          identity))
        except (SchemeError, ParserError), e:
            print "ERROR (%s): %s" % (e.__class__, e)
    p = Prompt(name="PyScheme", callback=promptCallback,
               quit_str=None)
    p.promptLoop()
            

if __name__ == '__main__':
    print 'Welcome to PyScheme!  Type: (QUIT) to quit.\n'
    repl()
