"""
Unit tests on PyScheme.
"""


import scheme
import unittest
import symbol
from symbol import Symbol
from error import SchemeError
import sys

parse = scheme.parse
eval = scheme.eval
pe = lambda s, e: eval(parse(s), e)  ## parse and eval


class SchemeTests(unittest.TestCase):
    def setUp(self):
         ## sets the resursion limit to something fairly small to test out the the tail recursion.
        sys.setrecursionlimit(30)
        self.env = scheme.setupEnvironment()


    def tearDown(self):
        self.env = None


    def pe(self, s):
        return pe(s, self.env)


    def testSimpleStuff(self):
        self.assertEquals(42, self.pe("42"))


    def testDefinition(self):
        self.assertRaises(SchemeError, self.pe, "pi") ## lookup before defn
        self.assertEquals(Symbol("ok"), self.pe("(define pi 3.1415926)"))
        self.assertEquals(3.1415926, self.pe("pi"))


    def testLet(self):
        self.assertEquals(5,
                          self.pe("""(let ((x 3)
                                           (y 4)
                                           (z 2))
                                        (+ x z))"""))

    def testAssignment(self):
        self.pe("(define x 0)")
        self.pe("""(define (inc*2)
                      (set! x (+ x 1))
                      (set! x (+ x 1)))""")
        self.pe("(inc*2)")
        self.assertEquals(2, self.pe("x"))



    def defineRecursiveFunctions(self):
        self.pe("""(define (factorial x)
                      (if (= x 0) 1 (* x (factorial (- x 1)))))""")

        self.pe("""(define (even? x) (if (= x 0) #t (odd? (- x 1))))""")
        self.pe("""(define (odd? x) (if (= x 0) #f (even? (- x 1))))""")

        
    def testRecursion(self):
        self.defineRecursiveFunctions()
        self.assertEquals(6, self.pe("(factorial 3)"))
        self.assertEquals(symbol.true, self.pe("(even? 4)"))


    def testTailRecursiveEval(self):
        """This is a test to see that eval() doesn't break, even on
        deeply recursive functions.  Basically, this is a test of the
        trampoline."""
        self.defineRecursiveFunctions()
        def myFactorial(n):
            return reduce(lambda x, y: x*y, range(1,n+1), 1)
        self.assertEquals(myFactorial(50), self.pe("(factorial 50)"))
        self.assertEquals(symbol.true, self.pe("(even? 42)"))



    def testArithmetic(self):
        self.assertEquals(3, self.pe("(+ 1 2)"))
        self.assertEquals(7, self.pe("(- 8 1)"))
        self.assertEquals(42, self.pe("(* 2 21)"))
        self.assertEquals(2, self.pe("(/ 4 2)"))



    def testIf(self):
        self.assertEquals(Symbol("danny"), self.pe("(if 42 'danny 'yoo)"))
        self.assertEquals(Symbol("yoo"), self.pe("(if #f 'danny 'yoo)"))



    def testCond(self):
        self.assertEquals(Symbol("ok"), self.pe(
            """(define (even? x) (= (remainder x 2) 0))"""))
        self.assertEquals(Symbol("ok"), self.pe(
            """(define (odd? x) (not (even? x)))"""))
        self.assertEquals(Symbol("ok"), self.pe(
            """(cond ((even? 3) "not-ok")
               ((odd? 2) "still-not-ok")
               (#t 'ok))"""))


    def testEvalCallcc(self):
        self.assertEquals(42, self.pe("(call/cc (lambda (x) 42))"))
        self.assertEquals(42, self.pe("(call/cc (lambda (x) (x 42)))"))
        self.assertEquals(15, self.pe("(+ 1 2 3 4 (call/cc (lambda (x) 5)))"))
        self.assertEquals(5, self.pe("(call/cc (lambda (x) (+ 1 2 3 4 (x 5))))"))
        self.pe("""
            (define (member x l)
               (call/cc (lambda (exit)
                            (cond ((null? l) (exit #f))
                                  ((equal? x (car l)) l)
                                  (#t (member x (cdr l)))))))""")
        self.assertEquals(self.pe("(list 4 5)"),
                          self.pe("(member 4 (list 1 2 3 4 5))"))


    def testLambda(self):
        pass

    def testBegin(self):
        pass

    def testApplication(self):
        pass
    

if __name__ == '__main__':
    unittest.main()
